from hashlib import sha3_256
from typing import Union

from nacl.signing import SigningKey, VerifyKey
from nacl.exceptions import BadSignatureError
from kasten import generator, Kasten
from kasten.exceptions import InvalidID
from kasten.types import KastenPacked, KastenChecksum

from onionrblocks.customtypes import RawEd25519PrivateKey, RawEd25519PublicKey

class Signed(generator.KastenBaseGenerator):

    @classmethod
    def generate(
            cls,
            packed_bytes: KastenPacked,
            signingKey: Union[SigningKey, RawEd25519PrivateKey]
            ) -> Kasten:
        """Sign a digest of packed bytes, return a Kasten instance of it."""
        # Use libsodium/pynacl (ed25519)
        hashed = sha3_256(packed_bytes).digest()
        try:
            signed = signingKey.sign(hashed)
        except AttributeError:
            signingKey = SigningKey(signingKey)
            signed = signingKey.sign(hashed)
        # The KastenChecksum will be 64 bytes message then 32 bytes of the hash
        # This can be fed right back into VerifyKey without splitting up
        return Kasten(signed, packed_bytes, None,
                      auto_check_generator=False)


    @staticmethod
    def validate_id(
            hash: KastenChecksum,
            packed_bytes: KastenPacked,
            verify_key: Union[VerifyKey, RawEd25519PublicKey]) -> None:
        # Hash is 64 bytes message then 32 bytes of the hash of the packed_bytes
        if len(hash) != 86:
            raise InvalidID("Block not have proper signature length")
        actual_hash = sha3_256(packed_bytes).digest()
        if not isinstance(verify_key, VerifyKey):
            verify_key = VerifyKey(verify_key)

        # Ensure that the digest is correct
        # Done in addition of the signature bc the sha3 can still ID blocks
        # and to prevent swapping sigs
        if actual_hash != hash[64:]:
            raise InvalidID("Invalid sha3_256 digest")
        # Ensure that the signature is correct
        try:
            verify_key.verify(hash)
        except BadSignatureError as e:
            raise InvalidID(repr(e))



from collections import namedtuple
from typing import NewType


AnonymousVDFBlockResult = namedtuple(
    'AnonymousVDFBlockResult', ['dec_checksum', 'raw_bytes'])

HexChecksum = NewType("HexChecksum", str)
DecChecksum = NewType("DecChecksum", int)
DecChecksumStr = NewType("DecChecksumStr", str)
RawEd25519PrivateKey = NewType("RawEd25519PrivateKey", bytes)
RawEd25519PublicKey = NewType("RawEd25519PublicKey", bytes)



from kasten import Kasten

from time import time
from math import floor
from typing import Iterable

import binascii

MAX_FUTURE_SKEW_SECONDS = 60 * 2
TOTAL_MAX_SIZE = 6 * (10**6)


class BlockRulesException(Exception):
    pass


class BlockExpired(BlockRulesException):
    pass

class BlockFutureSkewedBeyondMax(BlockRulesException):
    pass

class BlockExistsError(BlockRulesException):
    pass

class BlockTooLarge(BlockRulesException):
    pass


def checksum_exists_in_list(checksum: bytes, arr: 'Iterable'):
    if checksum in arr:
        raise BlockExistsError(
            binascii.hexlify(
                checksum).decode("utf-8") + " Already exists in given set")


def check_block_sanity(raw_bytes):
    kasten: Kasten = Kasten(None, raw_bytes, None, auto_check_generator=False)

    # Ensure a block is not too large
    if len(kasten.get_packed()) > TOTAL_MAX_SIZE:
        raise BlockTooLarge()

    # Ensure block is not future-skewed or expired
    timestamp: int = floor(time())
    block_timestamp = kasten.get_timestamp()
    difference = timestamp - block_timestamp

    if difference < -MAX_FUTURE_SKEW_SECONDS:
        raise BlockFutureSkewedBeyondMax()

    try:
        if block_timestamp + kasten.get_metadata()['expire'] >= timestamp:
            raise BlockExpired()
    except (IndexError, TypeError):
        pass





import time

from kasten import Kasten, generator
from kasten.types import KastenPacked, KastenChecksum
from kasten.exceptions import InvalidID

from mimcvdf import vdf_create, vdf_verify

from onionrblocks.universalrules import check_block_sanity


class AnonVDFGenerator(generator.KastenBaseGenerator):
    @classmethod
    def get_ttl_seconds_per_rounds(cls, rounds: int):
        # 8000 rounds = 1 second (2.8ghz python) = 1 hour storage
        if rounds < 8000:
            raise ValueError(
                "Rounds must be at least 8000")
        return (rounds / 8000) * 60

    @classmethod
    def generate(
            cls, packed_bytes: KastenPacked, rounds: int = 5000) -> Kasten:
        return Kasten(
            vdf_create(
                packed_bytes,
                rounds, dec=True
                ).to_bytes(
                    64, "big"), packed_bytes, cls, auto_check_generator=False)

    @staticmethod
    def validate_id(
            hash: KastenChecksum,
            packed_bytes: KastenPacked, rounds=5000) -> None:

        try:
            hash = int.from_bytes(hash, byteorder="big")
        except TypeError:
            pass
        if not vdf_verify(packed_bytes, hash, rounds):
            raise InvalidID

        test_obj = Kasten(
            None, packed_bytes, None, auto_check_generator=False)
        allowed_age_seconds = AnonVDFGenerator.get_ttl_seconds_per_rounds(
            rounds)
        if time.time() > test_obj.get_timestamp() + allowed_age_seconds:
            raise ValueError(
                f"Block rounds only vaild through {allowed_age_seconds}")

        return None